"""Common configuration defaults for the telemetry hub runtime."""

DEFAULT_STORAGE_PATH = "RTH_Store"
DEFAULT_ANNOUNCE_INTERVAL = 60
DEFAULT_HUB_TELEMETRY_INTERVAL = 600
DEFAULT_SERVICE_TELEMETRY_INTERVAL = 900
DEFAULT_LOG_LEVEL_NAME = "debug"
DEFAULT_MARKER_ANNOUNCE_INTERVAL_MINUTES = 15
DEFAULT_ANNOUNCE_CAPABILITIES_ENABLED = True
DEFAULT_ANNOUNCE_CAPABILITIES_MAX_BYTES = 256
DEFAULT_ANNOUNCE_CAPABILITIES_INCLUDE_VERSION = True
DEFAULT_ANNOUNCE_CAPABILITIES_INCLUDE_TIMESTAMP = False
